"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const path_1 = tslib_1.__importDefault(require("path"));
const fs_1 = tslib_1.__importDefault(require("fs"));
const os_1 = tslib_1.__importDefault(require("os"));
const tmp_1 = tslib_1.__importDefault(require("tmp"));
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const pify_1 = tslib_1.__importDefault(require("pify"));
const aws_sdk_1 = tslib_1.__importDefault(require("aws-sdk"));
const mkdirp_1 = tslib_1.__importDefault(require("mkdirp"));
const shelljs_1 = tslib_1.__importDefault(require("shelljs"));
const listr_1 = tslib_1.__importDefault(require("listr"));
const prompts_1 = require("./prompts");
const update_1 = require("./update");
const kyc_services_1 = require("./kyc-services");
const errors_1 = require("./errors");
const validate = tslib_1.__importStar(require("./validate"));
const utils = tslib_1.__importStar(require("./utils"));
const logger_1 = require("./logger");
const constants_1 = require("./constants");
tmp_1.default.setGracefulCleanup();
const mkdirp = pify_1.default(mkdirp_1.default);
const pfs = pify_1.default(fs_1.default);
const { prettify, isValidProjectPath, toEnvFile } = utils;
const silentLogger = Object.keys(logger_1.logger).reduce((silent, method) => {
    silent[method] = () => { };
    return silent;
}, {});
const AWS_CONF_PATH = `${os_1.default.homedir()}/.aws/config`;
const getFileNameForItem = item => `${item.id}.json`;
const read = file => fs_1.default.readFileSync(file, { encoding: 'utf8' });
const maybeRead = file => {
    if (fs_1.default.existsSync(file))
        return read(file);
};
const readJSON = file => JSON.parse(read(file));
const maybeReadJSON = file => {
    const result = maybeRead(file);
    if (result)
        return JSON.parse(result);
};
const write = (file, data) => fs_1.default.writeFileSync(file, prettify(data));
const pwrite = (file, data) => pfs.writeFile(file, prettify(data));
const exists = file => fs_1.default.existsSync(file);
const getLongFunctionName = ({ stackName, functionName }) => {
    if (functionName.lastIndexOf(stackName) === 0)
        return functionName;
    return `${stackName}-${functionName}`;
};
const readDirOfJSONs = dir => {
    return fs_1.default.readdirSync(dir)
        .map(file => require(path_1.default.resolve(dir, file)));
};
const normalizeError = err => {
    if (err instanceof Error)
        return err;
    const { name, message = 'unspecified' } = err;
    let normalized;
    if (name && name in global) {
        const ctor = global[name];
        try {
            normalized = new ctor(name);
        }
        catch (err) { }
    }
    if (!normalized) {
        normalized = new Error(message);
    }
    lodash_1.default.extend(normalized, err);
    return normalized;
};
const DEPLOYABLES = [
    'bot',
    'style',
    'models',
    'modelsPack',
    'terms'
];
const getDeployables = opts => lodash_1.default.pick(opts, DEPLOYABLES);
const getDeployablesKeys = opts => Object.keys(getDeployables(opts));
const DEPLOY_ALL_OPTS = DEPLOYABLES.reduce((obj, prop) => {
    obj[prop] = true;
    return obj;
}, {});
const getOptsOnly = opts => lodash_1.default.omit(opts, 'args');
const normalizeDeployOpts = (opts, command = 'deploy') => {
    if (!lodash_1.default.isEmpty(opts.args)) {
        throw new errors_1.Errors.InvalidInput(`unknown arguments: ${opts.args.join(' ')}`);
    }
    const all = opts.all || !lodash_1.default.size(getDeployables(opts));
    if (all) {
        return lodash_1.default.extend({}, DEPLOY_ALL_OPTS, opts);
    }
    if (!lodash_1.default.size(getDeployables(opts))) {
        throw new errors_1.Errors.InvalidInput(`you didn't indicate anything to ${command}!`);
    }
    return opts;
};
const functions = {
    setconf: 'setconf',
    cli: 'cli',
    importDataUtils: 'import_data_utils'
};
const paths = {
    conf: './conf',
    bot: './conf/bot.json',
    style: './conf/style.json',
    terms: './conf/terms-and-conditions.md',
    models: './models',
    lenses: './lenses'
};
read.bot = () => maybeReadJSON(paths.bot);
read.style = () => maybeReadJSON(paths.style);
read.models = () => readDirOfJSONs(paths.models);
read.lenses = () => readDirOfJSONs(paths.lenses);
read.terms = () => maybeRead(paths.terms);
const createImportDataUtilsMethod = ({ conf, method, props = [], required }) => (data) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    data = lodash_1.default.pick(data, props);
    (required || props).forEach(prop => {
        if (!(prop in data))
            throw new errors_1.Errors.InvalidInput(`expected "${prop}"`);
    });
    return yield conf.invokeAndReturn({
        functionName: functions.importDataUtils,
        arg: { method, data }
    });
});
class Conf {
    constructor(opts) {
        this.deploy = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const items = this.getDeployItems(opts);
            logger_1.logger.info(`deploying: ${getDeployablesKeys(items).join(', ')}`);
            if (opts.dryRun)
                return logger_1.logger.info('dry run, not executing');
            const { error, result } = yield this.invoke({
                functionName: functions.setconf,
                arg: items
            });
            if (error)
                throw error;
            return result;
        });
        this.invoke = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let { functionName, arg } = opts;
            if (!functionName)
                functionName = yield prompts_1.fn(this, 'which function?');
            let result;
            try {
                const promise = this.remote ? this._invoke(opts) : this._invokeLocal(opts);
                result = yield promise;
            }
            catch (error) {
                return { error };
            }
            return { result };
        });
        this.invokeAndReturn = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { error, result } = yield this.invoke(opts);
            if (error)
                throw error;
            return result;
        });
        this.getLongFunctionName = (functionName) => getLongFunctionName({
            stackName: this.stackName,
            functionName
        });
        this.getDeployItems = (opts) => {
            opts = normalizeDeployOpts(opts);
            const parts = {};
            if (opts.style) {
                parts.style = read.style();
            }
            if (opts.terms) {
                parts.terms = read.terms() || null;
            }
            if (opts.models) {
                const models = read.models();
                const lenses = read.lenses();
                if (models.length || lenses.length) {
                    parts.modelsPack = utils.pack({ models, lenses, namespace: this.namespace });
                }
            }
            if (opts.bot) {
                parts.bot = read.bot();
            }
            return parts;
        };
        this.load = (opts = {}) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const opLogger = opts.logger || logger_1.logger;
            opts = normalizeDeployOpts(opts, 'load');
            opLogger.info(`loading: ${getDeployablesKeys(opts).join(', ')}\n`);
            if (opts.dryRun)
                return opLogger.info('dry run, not executing');
            const result = yield this.exec({
                args: ['getconf --conf']
            });
            if (opts.style && result.style) {
                opLogger.debug('loaded remote style');
                write(paths.style, result.style);
            }
            if (opts.bot && result.bot) {
                opLogger.debug('loaded remote bot conf');
                write(paths.bot, result.bot);
            }
            if (opts.terms && result.termsAndConditions) {
                opLogger.debug('loaded remote terms and conditions');
                write(paths.terms, result.termsAndConditions.value);
            }
            if (opts.models && result.modelsPack) {
                opLogger.debug('loaded remote models and lenses');
                this.writeModels(result.modelsPack);
            }
        });
        this.writeModels = (modelsPack) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield ['models', 'lenses'].map((prop) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const arr = modelsPack[prop];
                if (!arr)
                    return;
                yield this.writeToFiles({
                    dir: paths[prop],
                    arr,
                    name: getFileNameForItem
                });
            }));
        });
        this.writeToFiles = ({ dir, name, arr }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield mkdirp(dir);
            yield Promise.all(arr.map(item => {
                return pwrite(path_1.default.join(dir, name(item)), item);
            }));
        });
        this.validate = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const items = this.getDeployItems(opts);
            lodash_1.default.each(items, (value, key) => {
                if (!validate[key]) {
                    logger_1.logger.info(`validation for ${key} will be done in cloud-side on deploy`);
                    return;
                }
                if (typeof value !== 'undefined') {
                    logger_1.logger.debug(`validating: ${key}`);
                    validate[key](value);
                }
            });
        });
        this.exec = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const res = yield this.invoke({
                functionName: functions.cli,
                arg: opts.args[0],
                noWarning: opts.noWarning
            });
            let { error, result } = res;
            if (error)
                throw normalizeError(error);
            if (!result) {
                throw new errors_1.Errors.ServerError('something went wrong, please wait a minute and try again');
            }
            ;
            ({ error, result } = result);
            if (error)
                throw normalizeError(error);
            return result;
        });
        this.init = (opts = {}) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { haveLocal, haveRemote, overwriteEnv, region, awsProfile, stack = {}, projectPath, loadCurrentConf, } = yield prompts_1.init(this);
            if (overwriteEnv === false)
                return;
            if (!(haveLocal || haveRemote)) {
                logger_1.logger.warn("Aborting. Re-run `tradleconf init` when you've either deployed a MyCloud, or have a local development environment, or both");
                return;
            }
            this.remote = haveRemote;
            this.region = region;
            this.profile = awsProfile;
            this.stackName = stack.name;
            this.stackId = stack.id;
            this.client = null;
            const env = utils.pickNonNull({
                region,
                awsProfile,
                stackName: this.stackName,
                stackId: this.stackId,
                project: projectPath && path_1.default.resolve(process.cwd(), projectPath)
            });
            const saveEnv = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                write('.env', toEnvFile(env));
                yield Promise.all([
                    paths.models,
                    paths.lenses,
                    paths.conf
                ].map(dir => mkdirp(dir)));
            });
            const saveEnvTask = {
                title: 'saving .env',
                task: saveEnv
            };
            if (haveRemote) {
                const tasks = [
                    {
                        title: 'loading deployment info',
                        task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                            ctx.info = yield this.getEndpointInfo();
                        })
                    },
                    {
                        title: 'initializing local conf',
                        task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                            env.apiBaseUrl = ctx.info.apiBaseUrl;
                            env.namespace = ctx.info.org.domain
                                .split('.')
                                .reverse()
                                .join('.');
                        })
                    },
                    saveEnvTask,
                ];
                if (loadCurrentConf) {
                    tasks.push({
                        title: 'loading remote configuration',
                        task: () => tslib_1.__awaiter(this, void 0, void 0, function* () { return this.load({ all: true, logger: silentLogger }); }),
                    });
                }
                yield new listr_1.default(tasks).run();
            }
            else {
                yield saveEnv();
            }
            logger_1.logger.success('initialization complete!');
        });
        this.createAWSClient = (opts = {}) => {
            const { profile = this.profile || process.env.awsProfile, region = this.region } = opts;
            if (region) {
                aws_sdk_1.default.config.update({ region });
            }
            if (profile) {
                aws_sdk_1.default.config.update({
                    credentials: new aws_sdk_1.default.SharedIniFileCredentials({ profile })
                });
            }
            const s3 = new aws_sdk_1.default.S3();
            const cloudformation = new aws_sdk_1.default.CloudFormation();
            const lambda = new aws_sdk_1.default.Lambda();
            const ecr = new aws_sdk_1.default.ECR();
            const ec2 = new aws_sdk_1.default.EC2();
            const opsworks = new aws_sdk_1.default.OpsWorks();
            return {
                s3,
                cloudformation,
                lambda,
                ecr,
                ec2,
                opsworks,
                region: aws_sdk_1.default.config.region,
            };
        };
        this.getStacks = (opts = {}) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const client = this.createAWSClient(opts);
            return yield utils.listStacks(client);
        });
        this.waitForStackUpdate = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { stackName = this.stackName } = opts || {};
            const client = this.createAWSClient();
            yield utils.awaitStackUpdate(client, stackName);
        });
        this.waitForStackDelete = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { stackName = this.stackName } = opts || {};
            const client = this.createAWSClient();
            yield utils.awaitStackDelete(this.client, stackName);
        });
        this.createDataBundle = ({ path }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let bundle;
            try {
                bundle = readJSON(path);
            }
            catch (err) {
                throw new errors_1.Errors.InvalidInput('expected "path" to bundle');
            }
            return yield this.invokeAndReturn({
                functionName: functions.importDataUtils,
                arg: {
                    method: 'createbundle',
                    data: bundle
                }
            });
        });
        this._createDataClaim = createImportDataUtilsMethod({
            conf: this,
            method: 'createclaim',
            props: ['key', 'claimType']
        });
        this.createDataClaim = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { qrCode } = opts;
            if (qrCode && !qrCode.endsWith('.png')) {
                throw new errors_1.Errors.InvalidInput(`expected qr code path to end with .png`);
            }
            const result = yield this._createDataClaim(opts);
            if (qrCode) {
                const { qrData } = result;
                yield utils.createQRCode(qrCode, qrData);
                logger_1.logger.info(`wrote qr code to: ${qrCode}\n`);
            }
            return result;
        });
        this.listDataClaims = createImportDataUtilsMethod({
            conf: this,
            method: 'listclaims',
            props: ['key']
        });
        this.getDataBundle = createImportDataUtilsMethod({
            conf: this,
            method: 'getbundle',
            props: ['key', 'claimId'],
            required: []
        });
        this._ensureRemote = () => {
            if (!this.remote) {
                throw new errors_1.Errors.InvalidInput('not supported for local dev env');
            }
        };
        this.destroy = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureStackNameKnown();
            this._ensureRemote();
            const { stackName } = this;
            yield prompts_1.confirmOrAbort(`DESTROY REMOTE MYCLOUD ${stackName}?? There's no undo for this one!`);
            yield prompts_1.confirmOrAbort(`Are you REALLY REALLY sure you want to MURDER ${stackName}?`);
            let buckets = yield utils.listStackBucketIds(this.client, stackName);
            buckets.forEach(id => logger_1.logger.info(id));
            yield prompts_1.confirmOrAbort('Delete these buckets?');
            const [big, small] = lodash_1.default.partition(buckets, id => constants_1.BIG_BUCKETS.find(logical => id.includes(logical.toLowerCase())));
            for (const id of small) {
                logger_1.logger.info(`emptying and deleting: ${id}`);
                yield utils.destroyBucket(this.client, id);
            }
            logger_1.logger.info('Note: it may take a few minutes for your stack to be deleted');
            yield new listr_1.default([
                {
                    title: 'disabling termination protection',
                    task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        yield utils.disableStackTerminationProtection(this.client, stackName);
                    })
                },
                {
                    title: 'deleting stack',
                    task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        yield utils.deleteStack(this.client, { StackName: stackName });
                        yield utils.wait(5000);
                        try {
                            yield this.waitForStackDelete();
                        }
                        catch (err) {
                            yield utils.deleteStack(this.client, { StackName: stackName, RetainResources: constants_1.BIG_BUCKETS });
                        }
                    })
                },
            ]).run();
            yield Promise.all(big.map(id => utils.markBucketForDeletion(this.client, id)));
            logger_1.logger.info(`The following buckets are too large to delete directly: ${big.join(', ')}.
Instead, I've marked them for deletion by S3. They should be gone within a day or so`);
        });
        this.getApiBaseUrl = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this.local) {
                return null;
            }
            return utils.getApiBaseUrl(this.client, this.stackName);
        });
        this.info = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureStackNameKnown();
            return yield this._info();
        });
        this._info = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const getLinks = this.invokeAndReturn({
                functionName: 'cli',
                arg: 'links',
                noWarning: true
            });
            const getInfo = this.remote ? this.getEndpointInfo() : Promise.resolve({});
            const [links, info] = yield Promise.all([getLinks, getInfo]);
            return Object.assign({ links: links.result }, lodash_1.default.pick(info, ['version', 'chainKey', 'apiBaseUrl']));
        });
        this.getEndpointInfo = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const getApiBaseUrl = this.getApiBaseUrl();
            const info = yield this.invokeAndReturn({
                functionName: 'info',
                arg: {},
                noWarning: true
            });
            if (info.statusCode !== 200) {
                throw new errors_1.Errors.ServerError(info.body);
            }
            if (info.isBase64Encoded) {
                info.body = new Buffer(info.body, 'base64');
            }
            const endpoint = JSON.parse(info.body);
            endpoint.apiBaseUrl = yield getApiBaseUrl;
            return endpoint;
        });
        this.getFunctions = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureStackNameKnown();
            const { client, stackName } = this;
            return yield utils.listStackFunctionIds(client, stackName);
        });
        this.getFunctionShortNames = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const functions = yield this.getFunctions();
            return functions.map(f => f.slice(this.stackName.length + 1));
        });
        this.tail = (opts = {}) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.log(Object.assign({ watch: true }, opts));
        });
        this.log = (opts = {}) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureStackNameKnown();
            this._ensureRemote();
            utils.checkCommandInPath('awslogs');
            const { client, stackName } = this;
            let functionName = opts.args[0];
            if (!functionName) {
                functionName = yield prompts_1.fn(this, 'which one do you want to log?');
            }
            const longName = getLongFunctionName({ stackName, functionName });
            const logOpts = getOptsOnly(opts);
            if (this.profile)
                logOpts.profile = this.profile;
            if (this.region)
                logOpts['aws-region'] = this.region;
            const passThrough = Object.keys(logOpts);
            const logOptsStr = passThrough
                .map(opt => {
                const key = utils.splitCamelCase(opt)
                    .join('-')
                    .toLowerCase();
                const val = logOpts[opt];
                if (val === true) {
                    return `--${key}`;
                }
                return `--${key}=${val}`;
            })
                .join(' ');
            const cmd = `awslogs get /aws/lambda/${longName} ${logOptsStr}`;
            logger_1.logger.info(cmd);
            shelljs_1.default.exec(cmd);
        });
        this.balance = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            return yield this.exec({
                args: ['balance']
            });
        });
        this.disable = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            return yield this.exec({
                args: ['setenvvar --key DISABLED --value 1']
            });
        });
        this.enable = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            return yield this.exec({
                args: ['setenvvar --key DISABLED']
            });
        });
        this.query = (query) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const queryString = JSON.stringify(query.replace(/[\s]/g, ' '));
            return yield this.exec({
                args: [`graphql --query ${queryString}`]
            });
        });
        this.getMyIdentity = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            return yield this.invokeAndReturn({
                functionName: 'cli',
                arg: 'identity',
                noWarning: true
            });
        });
        this.getMyPermalink = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const identity = yield this.getMyIdentity();
            return identity._permalink;
        });
        this.getCurrentVersion = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const versions = yield this.exec({
                args: ['listmyversions --limit 1'],
                noWarning: true
            });
            return versions[0];
        });
        this.listUpdates = ({ provider } = {}) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let command = 'listupdates';
            if (provider)
                command = `${command} --provider ${provider}`;
            return yield this.exec({
                args: [command],
                noWarning: true
            });
        });
        this.listPreviousVersions = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            return yield this.exec({
                args: [`listmyversions`],
                noWarning: true
            });
        });
        this.requestUpdate = ({ tag, provider }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!tag)
                throw new errors_1.Errors.InvalidInput('expected string "tag"');
            this._ensureRemote();
            let command = `getupdate --tag "${tag}"`;
            if (provider)
                command = `${command} --provider ${provider}`;
            yield this.exec({
                args: [command],
                noWarning: true
            });
        });
        this.getUpdateInfo = ({ tag }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            const result = yield this.exec({
                args: [`getupdateinfo --tag "${tag}"`],
                noWarning: true
            });
            if (!result.update) {
                throw new errors_1.Errors.NotFound(`not found: update with version: ${tag}`);
            }
            return result;
        });
        this.update = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            this._ensureStackNameKnown();
            yield update_1.update(this, Object.assign({ stackName: this.stackName }, opts));
        });
        this.updateManually = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.applyUpdateAsCurrentUser(opts);
            yield this.waitForStackUpdate();
        });
        this.rollback = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            this._ensureStackNameKnown();
            yield update_1.update(this, Object.assign({ stackName: this.stackName, rollback: true }, opts));
        });
        this.applyUpdateAsCurrentUser = (update) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            this._ensureStackNameKnown();
            const { templateUrl, notificationTopics } = update;
            const params = {
                StackName: this.stackId || this.stackName,
                TemplateURL: templateUrl,
                Capabilities: ['CAPABILITY_NAMED_IAM'],
            };
            if (notificationTopics) {
                params.NotificationARNs = notificationTopics;
            }
            yield utils.updateStack(this.client, params);
        });
        this.applyUpdateViaLambda = (update) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { templateUrl, notificationTopics } = update;
            return yield this.invokeAndReturn({
                functionName: 'updateStack',
                arg: { templateUrl, notificationTopics }
            });
        });
        this.enableKYCServices = (opts) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            this._ensureRegionKnown();
            yield kyc_services_1.configureKYCServicesStack(this, Object.assign({}, opts, { mycloudStackName: this.stackName, mycloudRegion: this.region, client: this.client }));
        });
        this.getPrivateConfBucket = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureStackNameKnown();
            const buckets = yield utils.listStackBucketIds(this.client, this.stackName);
            return buckets.find(bucket => /tdl-.*?-ltd-.*?-privateconfbucket/.test(bucket));
        });
        this.reboot = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            const functions = yield this.getFunctions();
            logger_1.logger.info(`rebooting functions:\n${functions.join('\n')}`);
            const DATE_UPDATED = String(Date.now());
            yield utils.updateEnvironments(this.client, {
                functions,
                transform: ({ name, env }) => (Object.assign({}, env, { DATE_UPDATED }))
            });
            logger_1.logger.info('warming up rebooted functions');
            yield this.warmup();
        });
        this.warmup = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._ensureRemote();
            return yield this.invokeAndReturn({
                functionName: 'genericJobRunner',
                arg: { name: 'warmup' },
                noWarning: true
            });
        });
        this._ensureStackNameKnown = () => {
            if (this.remote && !this.stackName) {
                throw new errors_1.Errors.InvalidInput(`hm...are you sure you're in the right directory?`);
            }
        };
        this._ensureRegionKnown = () => {
            if (this.remote && !this.region) {
                throw new errors_1.Errors.InvalidInput(`please re-run 'tradelconf init', your .env file is outdated`);
            }
        };
        this._invoke = ({ functionName, arg = {}, noWarning }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!(this.remote || noWarning)) {
                yield prompts_1.confirmOrAbort(`Targeting REMOTE deployment. Continue?`);
            }
            this._ensureStackNameKnown();
            const { StatusCode, Payload, FunctionError } = yield this.client.lambda.invoke({
                InvocationType: 'RequestResponse',
                FunctionName: this.getLongFunctionName(functionName),
                Payload: JSON.stringify(arg)
            }).promise();
            if (FunctionError || StatusCode >= 300) {
                const message = Payload || FunctionError;
                throw new Error(message.toString());
            }
            return JSON.parse(Payload.toString());
        });
        this._invokeLocal = ({ functionName, arg }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { project, nodeFlags } = this;
            const flagsStr = Object.keys(nodeFlags)
                .filter(key => nodeFlags[key])
                .map(key => `--${key}`)
                .join(' ');
            if (typeof arg !== 'string')
                arg = JSON.stringify(arg);
            const tmpInput = tmp_1.default.fileSync({ postfix: '.json' });
            const tmpOutput = tmp_1.default.fileSync({ postfix: '.json' });
            write(tmpInput.name, JSON.stringify(arg));
            const pwd = process.cwd();
            shelljs_1.default.cd(project);
            logger_1.logger.debug('be patient, local invocations can be slow');
            const envVars = lodash_1.default.pick(process.env, ['SERVERLESS_OFFLINE_APIGW']);
            if (!envVars.IS_OFFLINE)
                envVars.IS_OFFLINE = '1';
            const command = `${stringifyEnv(envVars)} node ${flagsStr} \
  "${project}/node_modules/.bin/sls" invoke local \
  -f "${functionName}" \
  -l false \
  --path "${tmpInput.name}" \
  --output "${tmpOutput.name}"`;
            logger_1.logger.debug(`running command: ${command}`);
            const result = shelljs_1.default.exec(command, { silent: true });
            shelljs_1.default.cd(pwd);
            const res = read(tmpOutput.name).trim();
            if (result.code !== 0)
                throw new Error(`invoke failed: ${res || result.stderr}`);
            return res && JSON.parse(res);
        });
        const { remote, region, profile, namespace, stackId, stackName, project, nodeFlags = {} } = opts;
        if (typeof remote !== 'boolean') {
            throw new errors_1.Errors.InvalidInput(`expected boolean "remote"`);
        }
        if (remote && !lodash_1.default.isEmpty(nodeFlags)) {
            throw new errors_1.Errors.InvalidInput('node debugging flags are only supported for local operations');
        }
        utils.normalizeNodeFlags(nodeFlags);
        this.nodeFlags = nodeFlags;
        this.region = region;
        this.namespace = namespace;
        this.profile = profile;
        this.stackName = stackName;
        this.stackId = stackId;
        this.project = project;
        this.remote = remote;
        this.local = !remote;
        let client;
        Object.defineProperty(this, 'client', {
            set(value) {
                client = value;
            },
            get() {
                if (!client) {
                    client = this.createAWSClient();
                }
                return client;
            }
        });
    }
}
exports.Conf = Conf;
exports.createConf = (opts) => new Conf(opts);
const stringifyEnv = props => Object.keys(props).map(key => `${key}="${props[key]}"`).join(' ');
//# sourceMappingURL=conf.js.map