"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const matches_1 = tslib_1.__importDefault(require("lodash/matches"));
const sortBy_1 = tslib_1.__importDefault(require("lodash/sortBy"));
const inquirer_1 = tslib_1.__importDefault(require("inquirer"));
const listr_1 = tslib_1.__importDefault(require("listr"));
const promise_retry_1 = tslib_1.__importDefault(require("promise-retry"));
const errors_1 = tslib_1.__importDefault(require("@tradle/errors"));
const errors_2 = require("./errors");
const logger_1 = require("./logger");
const prompts_1 = require("./prompts");
const USE_CURRENT_USER_ROLE = true;
const MIN_VERSION = '01.01.0f';
class Updater {
    constructor({ conf, opts, currentVersion, updates }) {
        this.update = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { targetTag, currentVersion, updates, opts } = this;
            const { provider, rollback } = opts;
            if (!currentVersion) {
                this.currentVersion = yield this.conf.getCurrentVersion();
            }
            if (!updates) {
                this.updates = yield this._loadUpdates();
            }
            if (targetTag && rollback) {
                const idx = this.updates.findIndex(matches_1.default({ tag: targetTag }));
                if (idx === -1) {
                    const previousTags = updates.map(u => u.tag);
                    throw new Error(`your MyCloud was never previously deployed with version ${targetTag}

Your previously deployed versions are (most recently deployed at the top):

${previousTags.join('\n')}`);
                }
            }
            yield this._update();
        });
        this._loadUpdates = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { conf, opts, currentVersion } = this;
            const { rollback, provider } = opts;
            const getUpdates = rollback
                ? conf.listPreviousVersions()
                : conf.listUpdates({ provider });
            let updates = yield getUpdates;
            updates = sortBy_1.default(updates, 'sortableTag');
            updates = updates.filter(u => u.tag !== currentVersion.tag);
            if (rollback) {
                updates = updates.reverse().filter(u => u.sortableTag < currentVersion.sortableTag);
            }
            return updates;
        });
        this._loadTargetTag = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { verb, updates, opts } = this;
            const { rollback, showReleaseCandidates } = opts;
            let choices = updates.slice();
            const noRC = !rollback && !showReleaseCandidates;
            if (noRC) {
                choices = choices.filter(update => !isReleaseCandidateTag(update.tag));
            }
            if (!choices.length) {
                throw new Error(`no ${verb} available`);
            }
            let message;
            if (rollback) {
                message = `Choose a version to roll back to (most recently deployed at the top)`;
            }
            else {
                message = `Choose a version to update to`;
            }
            const result = yield inquirer_1.default.prompt([{
                    type: 'rawlist',
                    pageSize: Infinity,
                    name: 'tag',
                    message,
                    choices: choices.map(({ tag }) => ({
                        name: getChoiceTextForTag(tag),
                        value: tag
                    })),
                }]);
            this.targetTag = result.tag;
        });
        this._update = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { conf, opts, currentVersion, updates, verb } = this;
            const { stackName, provider, showReleaseCandidates, force, rollback, } = opts;
            if (!(currentVersion.sortableTag && currentVersion.sortableTag > MIN_VERSION)) {
                throw new Error(`you have an old version of MyCloud which doesn't support the new update mechanism
  Please update manually this one time. See instructions on https://github.com/tradle/serverless`);
            }
            if (!this.targetTag) {
                yield this._loadTargetTag();
            }
            if (!rollback && currentVersion.templateUrl) {
                logger_1.logger.info(`
should the worst happen and you need to manually roll back to this version later, run:

tradleconf update-manually --template-url "${currentVersion.templateUrl}"
`);
            }
            const tag = this.targetTag;
            if (!force) {
                yield this._applyPrerequisiteTransitionTags();
            }
            const promise = yield new listr_1.default([
                {
                    title: `load release ${tag} (grab a coffee)`,
                    task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        const resp = yield this._getUpdateWithRetry();
                        if (!resp)
                            return;
                        const { update, upToDate } = resp;
                        ctx.upToDate = upToDate;
                        ctx.update = update;
                        ctx.willUpdate = force || rollback || !upToDate;
                    })
                },
                {
                    title: `validate release`,
                    skip: ctx => !ctx.willUpdate,
                    task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        try {
                            yield this._triggerUpdate(ctx.update);
                        }
                        catch (err) {
                            if (err.code === 'ValidationError' && err.message.includes('UPDATE_IN_PROGRESS')) {
                                throw new Error('stack is currently updating, please wait for it to finish before applying another update');
                            }
                            throw err;
                        }
                    })
                },
                {
                    title: `apply ${verb} (be patient, or else)`,
                    skip: ctx => !ctx.willUpdate,
                    task: (ctx) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        try {
                            yield conf.waitForStackUpdate({ stackName });
                        }
                        catch (err) {
                            if (err.code === 'ResourceNotReady') {
                                throw new Error(`failed to apply ${verb}`);
                            }
                            throw err;
                        }
                    })
                }
            ]).run();
            let ctx;
            try {
                ctx = yield promise;
            }
            catch (err) {
                if (errors_1.default.matches(err, errors_2.Errors.NotFound)) {
                    throw new Error(`failed to fetch version ${tag}. If you're confident it exists, please try again.`);
                }
                throw err;
            }
            if (ctx.upToDate && !ctx.willUpdate) {
                if (tag === currentVersion.tag) {
                    throw new Error(`your MyCloud is already up to date!`);
                }
                this._throwBackwardsError();
            }
        });
        this._throwBackwardsError = () => {
            const { currentVersion, targetTag } = this;
            throw new Error(`your MyCloud is of a more recent version: ${currentVersion.tag}

To force deploy ${targetTag}, run: tradleconf update --tag ${targetTag} --force`);
        };
        this._getUpdateWithRetry = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const tag = this.targetTag;
            const { conf, opts } = this;
            const { provider } = opts;
            let requested;
            return promise_retry_1.default((retry, number) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                try {
                    return yield conf.getUpdateInfo({ tag });
                }
                catch (err) {
                    errors_1.default.rethrow(err, ['developer', errors_2.Errors.InvalidInput]);
                    if (!requested && errors_1.default.matches(err, errors_2.Errors.NotFound)) {
                        requested = true;
                        yield conf.requestUpdate({ tag, provider });
                    }
                    retry(err);
                }
            }), {
                maxTimeout: 10000,
                minTimeout: 5000,
                retries: 10
            });
        });
        this._applyPrerequisiteTransitionTags = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { updates, currentVersion, targetTag } = this;
            const idx = updates.findIndex(update => update.tag === targetTag);
            if (idx === -1)
                return;
            const transition = updates.slice(0, idx).find(update => isTransitionReleaseTag(update.tag));
            if (!transition)
                return;
            logger_1.logger.info(`you must apply the transition version first: ${transition.tag}`);
            yield prompts_1.confirmOrAbort(`apply transition tag ${transition.tag} now?`);
            yield exports.update(this.conf, Object.assign({}, this.opts, { force: true, tag: transition.tag, currentVersion,
                updates }));
        });
        this._triggerUpdate = (update) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { conf } = this;
            if (USE_CURRENT_USER_ROLE) {
                yield conf.applyUpdateAsCurrentUser(update);
            }
            else {
                yield conf.applyUpdateViaLambda(update);
            }
        });
        this.conf = conf;
        this.opts = opts;
        this.currentVersion = currentVersion;
        this.updates = updates;
        this.targetTag = opts.tag;
        this.verb = opts.rollback ? 'rollback' : 'update';
    }
}
const isReleaseCandidateTag = (tag) => /-rc\.\d+$/.test(tag);
const isTransitionReleaseTag = (tag) => /-trans/.test(tag);
const getChoiceTextForTag = (tag) => {
    if (isTransitionReleaseTag(tag)) {
        return `${tag} (transition version, must be applied before the version that follows)`;
    }
    if (isReleaseCandidateTag(tag)) {
        return `${tag} (release candidate with experimental fixes / features)`;
    }
    return tag;
};
exports.update = (conf, opts) => new Updater({ conf, opts }).update();
//# sourceMappingURL=update.js.map