"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const util_1 = require("util");
const crypto_1 = tslib_1.__importDefault(require("crypto"));
const path_1 = tslib_1.__importDefault(require("path"));
const fs_1 = tslib_1.__importDefault(require("fs"));
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const shelljs_1 = tslib_1.__importDefault(require("shelljs"));
const node_fetch_1 = tslib_1.__importDefault(require("node-fetch"));
const aws_sdk_1 = tslib_1.__importDefault(require("aws-sdk"));
const models_pack_1 = tslib_1.__importDefault(require("@tradle/models-pack"));
const validate_resource_1 = tslib_1.__importDefault(require("@tradle/validate-resource"));
const errors_1 = tslib_1.__importDefault(require("@tradle/errors"));
const empty_bucket_1 = require("./empty-bucket");
const models_1 = require("./models");
const errors_2 = require("./errors");
const constants_1 = require("./constants");
const MY_CLOUD_STACK_NAME_REGEX = /^tdl-(.*?)-ltd-([a-zA-Z]+)$/;
exports.get = (url) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const res = yield node_fetch_1.default(url);
    if (res.statusCode > 300) {
        throw new Error(res.statusText);
    }
    return yield res.json();
});
exports.getNamespace = ({ models, lenses }) => {
    const model = lodash_1.default.values(models)[0] || lodash_1.default.values(lenses)[0];
    return model && models_pack_1.default.getNamespace(model.id);
};
exports.pack = ({ namespace, models, lenses }) => {
    if (!(lodash_1.default.size(models) || lodash_1.default.size(lenses))) {
        throw new errors_2.Errors.InvalidInput('expected "models" and/or "lenses"');
    }
    if (!namespace) {
        namespace = exports.getNamespace({ models, lenses });
    }
    const modelsPack = models_pack_1.default.pack({ namespace, models, lenses });
    return modelsPack;
};
exports.validateResource = resource => validate_resource_1.default({
    models: models_1.models,
    resource
});
exports.prettify = obj => {
    return typeof obj === 'string' ? obj : JSON.stringify(obj, null, 2);
};
exports.isValidProjectPath = project => {
    return fs_1.default.existsSync(path_1.default.resolve(project, 'serverless.yml'));
};
exports.toEnvFile = obj => Object.keys(obj)
    .map(key => `${key}="${obj[key]}"`)
    .join('\n');
const acceptAll = (item) => true;
exports.listStackResources = (aws, StackName, filter = acceptAll) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    let resources = [];
    const opts = { StackName };
    while (true) {
        let { StackResourceSummaries, NextToken } = yield aws.cloudformation.listStackResources(opts).promise();
        resources = resources.concat(StackResourceSummaries.filter(filter));
        opts.NextToken = NextToken;
        if (!opts.NextToken)
            break;
    }
    return resources;
});
exports.listStackResourcesByType = (type, aws, stackName) => {
    return exports.listStackResources(aws, stackName, ({ ResourceType }) => type === ResourceType);
};
exports.listStackResourceIdsByType = (type, aws, stackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const resources = yield exports.listStackResourcesByType(type, aws, stackName);
    return resources.map(r => r.PhysicalResourceId);
});
exports.listStackBuckets = exports.listStackResourcesByType.bind(null, 'AWS::S3::Bucket');
exports.listStackBucketIds = exports.listStackResourceIdsByType.bind(null, 'AWS::S3::Bucket');
exports.listStackFunctions = exports.listStackResourcesByType.bind(null, 'AWS::Lambda::Function');
exports.listStackFunctionIds = exports.listStackResourceIdsByType.bind(null, 'AWS::Lambda::Function');
exports.destroyBucket = (aws, Bucket) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    yield empty_bucket_1.emptyBucket(aws.s3, Bucket);
    try {
        yield aws.s3.deleteBucket({ Bucket }).promise();
    }
    catch (err) {
        errors_1.default.ignore(err, [
            { code: 'ResourceNotFoundException' },
            { code: 'NoSuchBucket' },
        ]);
    }
});
exports.markBucketForDeletion = (aws, Bucket) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const params = {
        Bucket,
        LifecycleConfiguration: {
            Rules: [
                {
                    Status: 'Enabled',
                    ID: 'expires-in-1-day',
                    Prefix: '',
                    Expiration: {
                        Days: 1,
                    },
                    NoncurrentVersionExpiration: {
                        NoncurrentDays: 1
                    }
                }
            ]
        }
    };
    yield aws.s3.putBucketLifecycleConfiguration(params).promise();
});
exports.disableStackTerminationProtection = (aws, StackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    return yield aws.cloudformation.updateTerminationProtection({
        StackName,
        EnableTerminationProtection: false,
    }).promise();
});
exports.deleteStack = (aws, params) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    while (true) {
        try {
            yield aws.cloudformation.deleteStack(params).promise();
            return () => exports.awaitStackDelete(aws, params.StackName);
        }
        catch (err) {
            if (!err.message.includes('UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS')) {
                throw err;
            }
            yield exports.wait(5000);
        }
    }
});
exports.createStackInRegion = ({ region, params }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    return exports.createStack({
        cloudformation: new aws_sdk_1.default.CloudFormation({ region })
    }, params);
});
exports.updateStackInRegion = ({ region, params }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    return exports.updateStack({
        cloudformation: new aws_sdk_1.default.CloudFormation({ region })
    }, params);
});
exports.createStack = ({ cloudformation }, params) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    yield cloudformation.createStack(params).promise();
    return () => exports.awaitStackCreate({ cloudformation }, params.StackName);
});
exports.updateStack = ({ cloudformation }, params) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    yield cloudformation.updateStack(params).promise();
    return () => exports.awaitStackUpdate({ cloudformation }, params.StackName);
});
const isUpdateableStatus = (status) => {
    return status.endsWith('_COMPLETE') && !status.startsWith('DELETE_');
};
exports.getStackId = (aws, stackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const stacks = yield exports.listStacks(aws, ({ StackName, StackStatus }) => StackName === stackName && isUpdateableStatus(StackStatus));
    return stacks[0] && stacks[0].id;
});
exports.awaitStackCreate = ({ cloudformation }, stackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    return exports.waitFor({ cloudformation, event: 'stackCreateComplete', stackName });
});
exports.awaitStackDelete = ({ cloudformation }, stackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    return exports.waitFor({ cloudformation, event: 'stackDeleteComplete', stackName });
});
exports.awaitStackUpdate = ({ cloudformation }, stackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    return exports.waitFor({ cloudformation, event: 'stackUpdateComplete', stackName });
});
exports.waitFor = ({ cloudformation, stackName, event }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    try {
        return yield cloudformation.waitFor(event, { StackName: stackName }).promise();
    }
    catch (err) {
        const url = exports.getConsoleLinkForStacksInRegion({ region: cloudformation.config.region, status: 'failed' });
        throw new errors_2.Errors.ServerError(`operation may have failed. Check your stacks here: ${url}`);
    }
});
exports.listStacks = (aws, filter = acceptAll) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const listStacksOpts = {
        StackStatusFilter: ['CREATE_COMPLETE', 'UPDATE_COMPLETE', 'UPDATE_ROLLBACK_COMPLETE']
    };
    let stackInfos = [];
    let keepGoing = true;
    while (keepGoing) {
        let { NextToken, StackSummaries } = yield aws.cloudformation.listStacks(listStacksOpts).promise();
        stackInfos = stackInfos.concat(StackSummaries.filter(filter).map(({ StackId, StackName }) => ({
            id: StackId,
            name: StackName
        })));
        listStacksOpts.NextToken = NextToken;
        keepGoing = !!NextToken;
    }
    return stackInfos;
});
exports.getApiBaseUrl = (aws, StackName) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const result = yield aws.cloudformation.describeStacks({ StackName }).promise();
    const { Outputs } = result.Stacks[0];
    const endpoint = Outputs.find(x => /^ServiceEndpoint/.test(x.OutputKey));
    return endpoint.OutputValue;
});
exports.splitCamelCase = str => str.split(/(?=[A-Z])/g);
exports.checkCommandInPath = cmd => {
    const { code } = shelljs_1.default.exec(`command -v ${cmd}`);
    if (code !== 0) {
        throw new errors_2.Errors.InvalidEnvironment(`Please install: ${cmd}`);
    }
};
exports.wait = millis => new Promise(resolve => setTimeout(resolve, millis));
exports.normalizeNodeFlags = flags => {
    if (!(flags.inspect || flags['inspect-brk'] || flags.debug || flags['debug-brk']))
        return;
    const nodeVersion = Number(process.version.slice(1, 2));
    if (nodeVersion >= 8) {
        if (flags.debug) {
            delete flags.debug;
            flags.inspect = true;
        }
        if (flags['debug-brk']) {
            delete flags['debug-brk'];
            flags['inspect-brk'] = true;
        }
    }
    else {
        if (flags.debug || flags['debug-brk']) {
            flags.inspect = true;
        }
    }
};
exports.pickNonNull = obj => lodash_1.default.pickBy(obj, val => val != null);
const sha256 = (str, enc = 'hex') => {
    return crypto_1.default.createHash('sha256').update(str).digest(enc);
};
const shortenString = (str, maxLength) => {
    const tooLongBy = str.length - maxLength;
    if (tooLongBy <= 0)
        return str;
    if (str.length < 6)
        throw new Error(`string is too short: ${str}`);
    return str.slice(0, maxLength - 6) + sha256(str).slice(0, 6);
};
const parseMyCloudStackName = (name) => {
    const [shortName, stage] = name.match(MY_CLOUD_STACK_NAME_REGEX).slice(1);
    return {
        shortName,
        stage,
    };
};
exports.isMyCloudStackName = (name) => MY_CLOUD_STACK_NAME_REGEX.test(name);
exports.getServicesStackName = (stackName) => {
    const { shortName } = parseMyCloudStackName(stackName);
    const shorterName = shortenString(shortName, 14);
    return `${shorterName}-srvcs`;
};
exports.canAccessECRRepos = (aws, { accountId, repoNames }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    try {
        yield aws.ecr.describeRepositories({
            registryId: accountId,
            repositoryNames: repoNames
        }).promise();
        return true;
    }
    catch (err) {
        if (err.name === 'AccessDeniedException') {
            return false;
        }
        throw err;
    }
});
exports.doKeyPairsExist = (aws, names) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const params = {
        KeyNames: names
    };
    try {
        yield aws.ec2.describeKeyPairs(params).promise();
        return true;
    }
    catch (err) {
        return false;
    }
});
exports.listKeyPairs = (aws) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const { KeyPairs } = yield aws.ec2.describeKeyPairs().promise();
    return KeyPairs.map(k => k.KeyName);
});
exports.listAZs = ({ region }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const ec2 = new aws_sdk_1.default.EC2({ region });
    const { AvailabilityZones } = yield ec2.describeAvailabilityZones({
        Filters: [
            {
                Name: 'region-name',
                Values: [region]
            }
        ]
    }).promise();
    return AvailabilityZones
        .filter(a => a.RegionName === region)
        .map(a => a.ZoneName);
});
exports.getUsedEIPCount = (aws) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const { Addresses } = yield aws.ec2.describeAddresses().promise();
    return Addresses.length;
});
exports.getConsoleLinkForStacksInRegion = ({ region, status = 'active' }) => {
    return `https://${region}.console.aws.amazon.com/cloudformation/home?region=${region}#/stacks?filter=${status}`;
};
exports.updateEnvironments = (aws, { functions, transform }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const confs = yield Promise.all(functions.map(FunctionName => aws.lambda.getFunctionConfiguration({ FunctionName }).promise()));
    yield Promise.all(functions.map((FunctionName, i) => tslib_1.__awaiter(this, void 0, void 0, function* () {
        const conf = confs[i];
        const Variables = transform({
            name: FunctionName,
            env: conf.Environment.Variables
        });
        yield aws.lambda.updateFunctionConfiguration({
            FunctionName,
            Environment: { Variables }
        }).promise();
    })));
});
exports.xor = (...args) => args.reduce((acc, next) => {
}, false);
const isLocal = ({ local, remote, project }) => {
    if (local && remote) {
        throw new errors_2.Errors.InvalidInput('expected "local" or "remote" but not both');
    }
    if (local) {
        if (!project) {
            throw new errors_2.Errors.InvalidInput('expected "project", the path to your local serverless project');
        }
        if (!exports.isValidProjectPath(project)) {
            throw new errors_2.Errors.InvalidInput('expected "project" to point to serverless project dir');
        }
        return true;
    }
    if (typeof remote === 'boolean') {
        return !remote;
    }
    return !!project;
};
exports.normalizeConfOpts = (opts) => {
    const local = isLocal(opts);
    return Object.assign({}, opts, { local, remote: !local });
};
exports.isRemoteOnlyCommand = commandName => constants_1.REMOTE_ONLY_COMMANDS.includes(commandName);
exports.isSafeRemoteCommand = commandName => constants_1.SAFE_REMOTE_COMMANDS.includes(commandName);
exports.createQRCode = (...args) => {
    const QRCode = require('qrcode');
    const toFile = util_1.promisify(QRCode.toFile.bind(QRCode));
    return toFile(...args);
};
//# sourceMappingURL=utils.js.map